/**
 * Copyright (C) 2008 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h> /* atoi, rand */
#include <string.h>
#include <glib/gprintf.h>

#include "ltop-media.h"
#include "ltop-player.h"
#include "ltop-image.h"

#ifdef HAVE_VIDEO
#include "ltop-video.h"
#endif

#include "debug.h"

enum {
    PROP_0,
    PROP_LOOP,
    PROP_RELOAD,
    PROP_SHUFFLE,
};

enum {
    NEXT_ITEM_SIGNAL,
    END_REACHED_SIGNAL,
    LAST_SIGNAL,
};

struct _LtopPlayerPrivate {
    const gchar *playlist_file;

    GHashTable *settings;
    GList      *playlist;

    gboolean    loop;
    gboolean    reload;
    gboolean    shuffle;
    gboolean    started;

    LtopMedia  *cur_media;
    LtopMedia  *next_media;

    GList      *cur_item;
};

static guint player_signals[LAST_SIGNAL] = { 0 };

static GdkColor black = { 0, 0, 0, 0 };

/* if a playlist hashtable doesn't have these keys, it is not valid */
static const gchar *to_check_keys[] = { "type", "uri", NULL };

static void ltop_player_class_init      (LtopPlayerClass *class);
static void ltop_player_init            (LtopPlayer *player);

static void ltop_player_finalize        (GObject *object);
static void ltop_player_set_property    (GObject *object,
                                         guint prop_id,
                                         const GValue *value,
                                         GParamSpec *pspec);
static void ltop_player_get_property    (GObject *object,
                                         guint prop_id,
                                         GValue *value,
                                         GParamSpec *pspec);
static void ltop_player_realize         (GtkWidget *widget);
static void ltop_player_size_request    (GtkWidget *widget,
                                         GtkRequisition *requisition);

/* helpers prototypes */
static gboolean playlist_read           (LtopPlayer *player);
static gboolean validate_item           (GHashTable *item);
static void playlist_free_item          (gpointer data,
                                         gpointer user_data);
static LtopMedia *ltop_player_get_media (LtopPlayer *player);

G_DEFINE_TYPE (LtopPlayer, ltop_player, GTK_TYPE_WIDGET);

static void
ltop_player_class_init (LtopPlayerClass *class)
{
    g_type_class_add_private (class, sizeof (LtopPlayerPrivate));

    GObjectClass *gobject_class = G_OBJECT_CLASS (class);
    gobject_class->set_property = ltop_player_set_property;
    gobject_class->get_property = ltop_player_get_property;
    gobject_class->finalize = ltop_player_finalize;

    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
    widget_class->realize = ltop_player_realize;
    widget_class->size_request = ltop_player_size_request;

    /* properties */
    g_object_class_install_property (gobject_class,
                                     PROP_LOOP,
                                     g_param_spec_boolean ("loop",
                                                           "loop",
                                                           "loop",
                                                           TRUE,
                                                           G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
                                     PROP_RELOAD,
                                     g_param_spec_boolean ("reload",
                                                           "reload",
                                                           "reload",
                                                           TRUE,
                                                           G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
                                     PROP_SHUFFLE,
                                     g_param_spec_boolean ("shuffle",
                                                           "shuffle",
                                                           "shuffle",
                                                           TRUE,
                                                           G_PARAM_READWRITE));

    /* signals */
    player_signals[NEXT_ITEM_SIGNAL] = g_signal_new (
                                  "next-item",
                                  G_TYPE_FROM_CLASS (class),
                                  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
                                  G_STRUCT_OFFSET (LtopPlayerClass, next_item),
                                  NULL,
                                  NULL,
                                  g_cclosure_marshal_VOID__VOID,
                                  G_TYPE_NONE, 0);

    player_signals[END_REACHED_SIGNAL] = g_signal_new (
                                  "end-reached",
                                  G_TYPE_FROM_CLASS (class),
                                  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
                                  G_STRUCT_OFFSET (LtopPlayerClass, end_reached),
                                  NULL,
                                  NULL,
                                  g_cclosure_marshal_VOID__VOID,
                                  G_TYPE_NONE, 0);

    /* Init the thread system */
    if (!g_thread_supported ())
        g_thread_init (NULL);
}

static void
ltop_player_init (LtopPlayer *player)
{
    player->priv = G_TYPE_INSTANCE_GET_PRIVATE (player,
                                                LTOP_TYPE_PLAYER,
                                                LtopPlayerPrivate);

    player->priv->settings = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                    g_free, g_free);
    player->priv->playlist = NULL;
    player->priv->next_media = NULL;
    player->priv->loop = TRUE;
    player->priv->reload = TRUE;
    player->priv->shuffle = FALSE;
    player->priv->started = FALSE;
}

static void
ltop_player_realize (GtkWidget *widget)
{
    DBG("entering");
    GdkWindowAttr attributes;
    gint attributes_mask;

    GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.event_mask = gtk_widget_get_events (widget);
    attributes.visual = gtk_widget_get_visual (widget);
    attributes.colormap = gtk_widget_get_colormap (widget);

    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
                                     &attributes, attributes_mask);

    widget->style = gtk_style_attach (widget->style, widget->window);

    gdk_window_set_user_data (widget->window, widget);

    gtk_style_set_background (widget->style, widget->window, GTK_STATE_ACTIVE);

    gdk_window_set_background (widget->window, &black);
    DBG("leaving");
}

static void
ltop_player_size_request (GtkWidget *widget,
                          GtkRequisition *requisition)
{
    /* FIXME: don't assume we use fullscreen */
    GdkScreen *screen;
    int h, w;

    screen = gdk_screen_get_default ();
    w = gdk_screen_get_width (screen);
    h = gdk_screen_get_height (screen);

    requisition->width = w;
    requisition->height = h;
}

static void
ltop_player_finalize (GObject *object)
{
    LtopPlayer *player = LTOP_PLAYER (object);

    if (LTOP_IS_MEDIA (player->priv->cur_media))
        g_object_unref (G_OBJECT (player->priv->cur_media));
    if (LTOP_IS_MEDIA (player->priv->next_media))
        g_object_unref (G_OBJECT (player->priv->next_media));

    g_hash_table_destroy (player->priv->settings);
    g_list_foreach (player->priv->playlist, playlist_free_item, NULL);
    g_list_free (player->priv->playlist);

    (*G_OBJECT_CLASS (ltop_player_parent_class)->finalize) (object);
}

static void
ltop_player_set_property (GObject *object,
                          guint prop_id,
                          const GValue *value,
                          GParamSpec *pspec)
{
    LtopPlayer *player = LTOP_PLAYER (object);

    switch (prop_id)
    {
        case PROP_LOOP:
            player->priv->loop = g_value_get_boolean (value);
            break;
        case PROP_RELOAD:
            player->priv->reload = g_value_get_boolean (value);
            break;
        case PROP_SHUFFLE:
            player->priv->shuffle = g_value_get_boolean (value);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}

static void
ltop_player_get_property (GObject *object,
                          guint prop_id,
                          GValue *value,
                          GParamSpec *pspec)
{
    LtopPlayer *player = LTOP_PLAYER (object);

    switch (prop_id)
    {
        case PROP_LOOP:
            g_value_set_boolean (value, player->priv->loop);
            break;
        case PROP_RELOAD:
            g_value_set_boolean (value, player->priv->reload);
            break;
        case PROP_SHUFFLE:
            g_value_set_boolean (value, player->priv->shuffle);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}

/* helpers */
static gboolean
validate_item (GHashTable *item)
{
    gboolean valid = TRUE;
    int i = 0;
    while (to_check_keys[i])
    {
        if (!g_hash_table_lookup (item, to_check_keys[i]))
        {
            valid = FALSE;
            break;
        }
        i++;
    }
    return valid;
}

static void
duplicate_item_table (gpointer key, gpointer value, gpointer user_data)
{
    GHashTable *ntable = (GHashTable *) user_data;
    /* we've already set uri and type */
    if (strcmp ((const gchar *) key, "type")
        && strcmp ((const gchar *) key, "uri"))
    {
        /* duplicate the values so we can free the orig hash table */
        g_hash_table_insert (ntable, g_strdup ((const gchar *) key),
                                     g_strdup ((const gchar *) value));
    }
}

static gint
shuffle_sorting_func (gconstpointer a,
                      gconstpointer b)
{
    gint ret = (rand () > (RAND_MAX / 2)) ? -1 : 1;
    DBG("random: %d", ret);
    return ret;
}

static void
playlist_free_item (gpointer data, gpointer user_data)
{
    GHashTable *table = (GHashTable *) data;
    g_hash_table_destroy (table);
}

static LtopMedia *
ltop_player_get_media (LtopPlayer *player)
{
    GHashTable *table = player->priv->cur_item->data;
    GHashTable *settings = player->priv->settings;
    gchar *type = NULL;
    gchar *uri = NULL;

    type = (gchar *) g_hash_table_lookup (table, "type");
    uri = (gchar *) g_hash_table_lookup (table, "uri");

#ifdef HAVE_VIDEO
    if (!strcmp (type, "video"))
    {
        DBG("new video media");
        return LTOP_MEDIA (ltop_video_new (uri));
    }
#endif

    if (!strcmp (type, "image"))
    {
        gchar *timeout_str = NULL;
        int timeout = 0;
        gchar *style_str = NULL;
        gint style = LTOP_IMAGE_STYLE_CENTERED;

        timeout_str = (gchar *) g_hash_table_lookup (table, "timeout");
        if (!timeout_str)
            timeout_str = (gchar *) g_hash_table_lookup (settings, "DefaultTimeout");
        if (timeout_str)
            timeout = atoi (timeout_str);

        style_str = (gchar *) g_hash_table_lookup (table, "style");
        if (!style_str)
            style_str = (gchar *) g_hash_table_lookup (settings, "DefaultImageStyle");
        if (style_str)
        {
            if (!strcmp (style_str, "tiled"))
                style = LTOP_IMAGE_STYLE_TILED;
            else if (!strcmp (style_str, "stretched"))
                style = LTOP_IMAGE_STYLE_STRETCHED;
            else if (!strcmp (style_str, "scaled"))
                style = LTOP_IMAGE_STYLE_SCALED;
        }

        DBG("new image media");
        return LTOP_MEDIA (ltop_image_new (uri, timeout, style));
    }

    return NULL;
}

/* xml parsing */
enum
{
    START = 0,
    SETTINGS,
    PLAYLIST,
    UNKNOWN
};

int parser_state = START;

static void
start_element_handler (GMarkupParseContext *context,
                       const gchar         *element_name,
                       const gchar        **attribute_names,
                       const gchar        **attribute_values,
                       gpointer             user_data,
                       GError             **error)
{
    LtopPlayer *player = LTOP_PLAYER (user_data);
    int i = 0;

    switch (parser_state)
    {
        case START:
        if (!strcmp (element_name, "settings"))
        {
            parser_state = SETTINGS;
        }
        else if (!strcmp (element_name, "playlist"))
        {
            parser_state = PLAYLIST;
        }
        break;

        case SETTINGS:
        if (!strcmp (element_name, "option"))
        {
            gchar *name = NULL;
            gchar *value = NULL;
            while (attribute_names[i] != NULL)
            {
                if (!strcmp (attribute_names[i], "name"))
                {
                    if (name)
                    {
                        g_free (name);
                        name = NULL;
                    }
                    name = g_strdup (attribute_values[i]);
                }
                else if (!strcmp (attribute_names[i], "value"))
                {
                    if (value)
                    {
                        g_free (value);
                        value = NULL;
                    }
                    value = g_strdup (attribute_values[i]);
                }
                i++;
            }
            /* name and value must not be freed here */
            if (name && strcmp (name, "")
                && value && strcmp (value, ""))
                    g_hash_table_insert (player->priv->settings, name, value);
        }
        break;

        case PLAYLIST:
        if (!strcmp (element_name, "item"))
        {
            GHashTable *table = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                       g_free, g_free);

            while (attribute_names[i] != NULL)
            {
                g_hash_table_insert (table,
                                     g_strdup (attribute_names[i]),
                                     g_strdup (attribute_values[i]));
                i++;
            }

            if (g_hash_table_size (table) > 0 && validate_item (table))
            {
                /* if we have an imagedir, we need to express it as a list of image items */
                gchar *type = NULL;
                type = (gchar *) g_hash_table_lookup (table, "type");
#ifdef HAVE_VIDEO
                if (!strcmp (type, "imagedir") || !strcmp (type, "videodir"))
#else
                if (!strcmp (type, "imagedir"))
#endif
                {
                    gchar *uri = NULL;
                    uri = (gchar *) g_hash_table_lookup (table, "uri");
                    GDir *dir = g_dir_open (uri, 0, NULL);
                    if (dir)
                    {
                        const gchar *file = NULL;
                        while ((file = g_dir_read_name (dir)) != NULL)
                        {
                            DBG("adding new item from dir: %s/%s", uri, file);
                            gchar *filename = g_build_filename (uri, file, NULL);
                            GHashTable *ntable = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                                        g_free, g_free);

                            if (!strcmp (type, "imagedir"))
                            {
                                g_hash_table_insert (ntable,
                                                     g_strdup ("type"),
                                                     g_strdup ("image"));
                            }
#ifdef HAVE_VIDEO
                            else if (!strcmp (type, "videodir")) {
                                g_hash_table_insert (ntable,
                                                     g_strdup ("type"),
                                                     g_strdup ("video"));
                            }
#endif

                            g_hash_table_insert (ntable,
                                                 g_strdup ("uri"),
                                                 g_strdup (filename));

                            g_hash_table_foreach (table, duplicate_item_table, (gpointer) ntable);

                            /* we know we have a uri and a type, so the new item is valid */
                            player->priv->playlist = g_list_append (player->priv->playlist, ntable);
                            g_free (filename);
                        }

                        g_dir_close (dir);
                    }
                    /* we've duplicated everything, we can free the orig table */
                    g_hash_table_destroy (table);
                } else {
                    DBG("adding an item of type %s with uri %s",
                        (gchar *)g_hash_table_lookup (table, "type"),
                        (gchar *)g_hash_table_lookup (table, "uri"));
                    player->priv->playlist = g_list_append (player->priv->playlist, table);
                }
            } else {
                g_hash_table_destroy (table);
            }
        }
        break;
    }

}

static void
end_element_handler (GMarkupParseContext *context,
                     const gchar         *element_name,
                     gpointer             user_data,
                     GError             **error)
{
    if ((parser_state == SETTINGS
            && !strcmp (element_name, "settings"))
        || (parser_state == PLAYLIST
            && !strcmp (element_name, "playlist")))
    {
        parser_state = START;
    }
}

static GMarkupParser markup_parser = {
    start_element_handler,
    end_element_handler,
    NULL,
    NULL,
    NULL
};

static gboolean
playlist_read (LtopPlayer *player)
{
    DBG("reading playlist");
    if (player->priv->playlist)
    {
        g_list_foreach (player->priv->playlist, playlist_free_item, NULL);
        g_list_free (player->priv->playlist);
        player->priv->playlist = NULL;
    }
    g_hash_table_remove_all (player->priv->settings);

    GMarkupParseContext *context;
    gchar *contents = NULL;
    size_t size;
    gboolean retval = TRUE;

    if (!g_file_get_contents (player->priv->playlist_file, &contents, &size, NULL))
    {
        g_fprintf (stderr,
                   "Unable to read the playlist contents.\n");
        return (FALSE);
    }

    context = g_markup_parse_context_new (&markup_parser, 0, player, NULL);
    if (!g_markup_parse_context_parse (context, contents, size, NULL))
    {
        g_fprintf (stderr,
                   "Error parsing the xml file.\n");
        retval = FALSE;
        goto cleanup;
    }
    if (!g_markup_parse_context_end_parse (context, NULL))
    {
        g_fprintf (stderr,
                   "Error parsing the xml file.\n");
        retval = FALSE;
        goto cleanup;
    }

cleanup:
    if (contents)
        g_free (contents);
    g_markup_parse_context_free (context);

    if (player->priv->shuffle)
        player->priv->playlist = g_list_sort (player->priv->playlist, shuffle_sorting_func);

    DBG("generated a playlist of %d elements", g_list_length (player->priv->playlist));
    return retval;
}

/* callbacks */
static gpointer
media_run (gpointer data)
{
    ltop_media_run (LTOP_MEDIA (data));
    return 0;
}

static void
on_media_job_done (LtopMedia *media, gpointer data)
{
    LtopPlayer *player = LTOP_PLAYER (data);

    gdk_threads_enter ();
    ltop_player_play_next (player);
    gdk_threads_leave ();
}

/* Public API */
GtkWidget *
ltop_player_new (const gchar *filename)
{
    DBG("entering");
    g_return_val_if_fail (filename != NULL, NULL);

    gboolean ok;
    LtopPlayer *player = g_object_new (ltop_player_get_type (), NULL);
    player->priv->playlist_file = filename;
    ok = playlist_read (player);

    if (ok)
    {
        return GTK_WIDGET (player);
    } else {
        return NULL;
    }
    DBG("leaving");
}

void
ltop_player_run (LtopPlayer *player)
{
    DBG("entering");
    ltop_player_play_next (player);
    DBG("leaving");
}

void
ltop_player_shuffle (LtopPlayer *player)
{
    player->priv->shuffle = TRUE;

    if (player->priv->shuffle)
        player->priv->playlist = g_list_sort (player->priv->playlist,
                                              shuffle_sorting_func);
}

void
ltop_player_play_next (LtopPlayer *player)
{
    DBG("playing next item");

    if (player->priv->cur_media != NULL)
    {
        DBG ("releasing old media");
        g_object_unref (player->priv->cur_media);
        player->priv->cur_media = NULL;
    }

    /* TODO (Jannis:) not sure this belongs into this function */
    if (!player->priv->started)
        player->priv->started = TRUE;

    if (player->priv->cur_item == NULL)
    {
        if (player->priv->reload) 
            playlist_read (player);

        player->priv->cur_item = g_list_first (player->priv->playlist);
    } else if (g_list_next (player->priv->cur_item) == NULL) {
        DBG ("end of playlist reached");
        
        if (player->priv->loop)
        {
            DBG ("playlist loop, going back to the first item");
            
            if (player->priv->reload) 
                playlist_read (player);

            player->priv->cur_item = g_list_first (player->priv->playlist);
        } else {
            DBG ("not looping, emitting end reached signal");
            g_signal_emit (player, player_signals[END_REACHED_SIGNAL], 0);
            player->priv->cur_item = NULL;
        }
    } else {
        player->priv->cur_item = g_list_next (player->priv->cur_item);
    }

    if (player->priv->cur_item != NULL)
    {
        DBG ("loading next media");

        player->priv->cur_media = ltop_player_get_media (player);

        g_signal_emit (G_OBJECT (player), player_signals[NEXT_ITEM_SIGNAL], 0);

        g_signal_connect (G_OBJECT (player->priv->cur_media), "job-done",
                          G_CALLBACK (on_media_job_done), (gpointer) player);

        ltop_media_attach_to_window (player->priv->cur_media, GTK_WIDGET (player)->window);

        /* run the media in a thread, because the video gobject waits for its
         * own thread (thus locks the player) */
        g_thread_create (media_run, (gpointer) player->priv->cur_media,
                         FALSE, NULL);
    }
}

